@@ -1,15 +1,32 @@
 /*
-  copyright : GPL
-    (author: joost witteveen, joostje@debian.org)
+  Copyright Ⓒ 1997, 1998, 1999, 2000, 2001  joost witteveen
+  Copyright Ⓒ 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009  Clint Adams
 
+    This program is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
 
- This file contains the code (wrapper functions) that gets linked with 
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+ This file contains the code (wrapper functions) that gets linked with
  the programes run from inside fakeroot. These programes then communicate
- with the fakeroot daemon, that keeps information about the "fake" 
+ with the fakeroot daemon, that keeps information about the "fake"
  ownerships etc. of the files etc.
-    
+
  */
 
+#ifdef __APPLE__
+/*
+   In this file, we want 'struct stat' to have a 32-bit 'ino_t'.
+   We use 'struct stat64' when we need a 64-bit 'ino_t'.
+*/
+#define _DARWIN_NO_64_BIT_INODE
+#endif
+
 #include "communicate.h"
 #include <dlfcn.h>
 #include <stdio.h>
@@ -71,9 +88,9 @@ static void fail(const char *msg)
 
 const char *env_var_set(const char *env){
   const char *s;
-  
+
   s=getenv(env);
-  
+
   if(s && *s)
     return s;
   else
@@ -85,7 +102,7 @@ void cpyfakemstat(struct fake_msg *f, const struct stat *st
 		, int ver
 #endif
 		){
-  
+
 #ifndef STUPID_ALPHA_HACK
   f->st.mode =st->st_mode;
   f->st.ino  =st->st_ino ;
@@ -338,7 +355,7 @@ void cpyfakefake(struct fakestat *dest,
 
 void stat64from32(struct stat64 *s64, const struct stat *s32)
 {
-  /* I've added st_size and st_blocks here. 
+  /* I've added st_size and st_blocks here.
      Don't know why they were missing -- joost*/
    s64->st_dev = s32->st_dev;
    s64->st_ino = s32->st_ino;
@@ -459,8 +476,13 @@ static void open_comm_sd(void)
   if (fcntl(comm_sd, F_SETFD, FD_CLOEXEC) < 0)
     fail("fcntl(F_SETFD, FD_CLOEXEC)");
 
-  if (connect(comm_sd, get_addr(), sizeof (struct sockaddr_in)) < 0)
-    fail("connect");
+  while (1) {
+    if (connect(comm_sd, get_addr(), sizeof (struct sockaddr_in)) < 0) {
+      if (errno != EINTR)
+        fail("connect");
+    } else
+      break;
+  }
 }
 
 void lock_comm_sd(void)
@@ -483,8 +505,10 @@ void send_fakem(const struct fake_msg *buf)
 
   if(init_get_msg()!=-1){
     ((struct fake_msg *)buf)->mtype=1;
-    r=msgsnd(msg_snd, (struct fake_msg *)buf,
-	     sizeof(*buf)-sizeof(buf->mtype), 0);
+    do
+      r=msgsnd(msg_snd, (struct fake_msg *)buf,
+	       sizeof(*buf)-sizeof(buf->mtype), 0);
+    while((r==-1) && (errno==EINTR));
     if(r==-1)
       perror("libfakeroot, when sending message");
   }
@@ -531,7 +555,13 @@ void send_get_fakem(struct fake_msg *buf)
       l=msgrcv(msg_get,
                (struct my_msgbuf*)buf,
                sizeof(*buf)-sizeof(buf->mtype),0,0);
-    while((buf->serial!=serial)||buf->pid!=pid);
+    while(((l==-1)&&(errno==EINTR))||(buf->serial!=serial)||buf->pid!=pid);
+
+    if(l==-1){
+      buf->xattr.flags_rc=errno;
+      fprintf(stderr,"fakeroot internal error #%d: %s\n",
+              errno, strerror(errno));
+    }
 
     semaphore_down();
 
@@ -572,7 +602,7 @@ static size_t read_all(int fd,void *buf,size_t count) {
 		  else fail("partial read");
 	  } else {
 		  remaining-=rc;
-	  }  
+	  }
   }
   return count-remaining;
 }
@@ -590,6 +620,9 @@ static void send_fakem_nr(const struct fake_msg *buf)
   fm.st.mode = htonl(buf->st.mode);
   fm.st.nlink = htonl(buf->st.nlink);
   fm.remote = htonl(0);
+  fm.xattr.buffersize = htonl(buf->xattr.buffersize);
+  fm.xattr.flags_rc = htonl(buf->xattr.flags_rc);
+  memcpy(fm.xattr.buf, buf->xattr.buf, MAX_IPC_BUFFER_SIZE);
 
   while (1) {
     ssize_t len;
@@ -636,7 +669,7 @@ static void get_fakem_nr(struct fake_msg *buf)
       continue;
    fail("read");
   }
- 
+
   buf->id = ntohl(buf->id);
   buf->st.uid = ntohl(buf->st.uid);
   buf->st.gid = ntohl(buf->st.gid);
@@ -646,6 +679,8 @@ static void get_fakem_nr(struct fake_msg *buf)
   buf->st.mode = ntohl(buf->st.mode);
   buf->st.nlink = ntohl(buf->st.nlink);
   buf->remote = ntohl(buf->remote);
+  buf->xattr.buffersize = ntohl(buf->xattr.buffersize);
+  buf->xattr.flags_rc = ntohl(buf->xattr.flags_rc);
 }
 
 void send_get_fakem(struct fake_msg *buf)
@@ -734,6 +769,63 @@ void send_get_stat(struct stat *st
   }
 }
 
+void send_get_xattr(struct stat *st
+		, xattr_args *xattr
+#ifdef STUPID_ALPHA_HACK
+		, int ver
+#endif
+		){
+  struct fake_msg buf;
+  size_t in_size;
+  size_t name_size;
+  size_t total_size;
+
+#ifndef FAKEROOT_FAKENET
+  if(init_get_msg()!=-1)
+#endif /* ! FAKEROOT_FAKENET */
+  {
+#ifndef STUPID_ALPHA_HACK
+    cpyfakemstat(&buf,st);
+#else
+    cpyfakemstat(&buf,st,ver);
+#endif
+    in_size = xattr->size;
+    total_size = (xattr->func == setxattr_func) ? (in_size) : 0;
+    if (xattr->name)
+    {
+      name_size = strlen(xattr->name);
+      total_size += name_size + 1;
+    }
+    if (total_size > MAX_IPC_BUFFER_SIZE)
+    {
+      xattr->rc = ERANGE;
+      return;
+    }
+    if (xattr->name) {
+      strcpy(buf.xattr.buf, xattr->name);
+      if (xattr->func == setxattr_func)
+        memcpy(&buf.xattr.buf[name_size + 1], xattr->value, in_size);
+    }
+    buf.xattr.buffersize = total_size;
+    buf.xattr.flags_rc = xattr->flags;
+    buf.id=xattr->func;
+    send_get_fakem(&buf);
+    xattr->rc = buf.xattr.flags_rc;
+    xattr->size = buf.xattr.buffersize;
+    if (buf.xattr.buffersize) {
+      if (!in_size) {
+        /* Special case. Return size of required buffer */
+        return;
+      }
+      if (xattr->size > in_size) {
+        xattr->rc = ERANGE;
+        return;
+      }
+      memcpy(xattr->value, buf.xattr.buf, xattr->size);
+    }
+  }
+}
+
 #ifdef STAT64_SUPPORT
 void send_get_stat64(struct stat64 *st
 #ifdef STUPID_ALPHA_HACK
@@ -762,6 +854,63 @@ void send_get_stat64(struct stat64 *st
 #endif
   }
 }
+
+void send_get_xattr64(struct stat64 *st
+		, xattr_args *xattr
+#ifdef STUPID_ALPHA_HACK
+		, int ver
+#endif
+		){
+  struct fake_msg buf;
+  size_t in_size;
+  size_t name_size;
+  size_t total_size;
+
+#ifndef FAKEROOT_FAKENET
+  if(init_get_msg()!=-1)
+#endif /* ! FAKEROOT_FAKENET */
+  {
+#ifndef STUPID_ALPHA_HACK
+    cpyfakemstat64(&buf,st);
+#else
+    cpyfakemstat64(&buf,st,ver);
+#endif
+    in_size = xattr->size;
+    total_size = (xattr->func == setxattr_func) ? (in_size) : 0;
+    if (xattr->name)
+    {
+      name_size = strlen(xattr->name);
+      total_size += name_size + 1;
+    }
+    if (total_size > MAX_IPC_BUFFER_SIZE)
+    {
+      xattr->rc = ERANGE;
+      return;
+    }
+    if (xattr->name) {
+      strcpy(buf.xattr.buf, xattr->name);
+      if (xattr->func == setxattr_func)
+        memcpy(&buf.xattr.buf[name_size + 1], xattr->value, in_size);
+    }
+    buf.xattr.buffersize = total_size;
+    buf.xattr.flags_rc = xattr->flags;
+    buf.id=xattr->func;
+    send_get_fakem(&buf);
+    xattr->rc = buf.xattr.flags_rc;
+    xattr->size = buf.xattr.buffersize;
+    if (buf.xattr.buffersize) {
+      if (!in_size) {
+        /* Special case. Return size of required buffer */
+        return;
+      }
+      if (xattr->size > in_size) {
+        xattr->rc = ERANGE;
+        return;
+      }
+      memcpy(xattr->value, buf.xattr.buf, xattr->size);
+    }
+  }
+}
 #endif /* STAT64_SUPPORT */
 
 #ifndef FAKEROOT_FAKENET
@@ -785,7 +934,7 @@ key_t get_ipc_key(key_t new_key)
 
 int init_get_msg(){
   /* a msgget call generates a fstat() call. As fstat() is wrapped,
-     that call will in turn call semaphore_up(). So, before 
+     that call will in turn call semaphore_up(). So, before
      the semaphores are setup, we should make sure we already have
      the msg_get and msg_set id.
      This is why semaphore_up() calls this function.*/
